#include "queue_internal.h"





/****************************************************************************/
int queue_enqueue( struct queue_state *qs, void *user_data )
{
  ALIGN(ALIGN_DOUBLE_POINTER) struct queue_element
    *volatile qe[QUEUE_PAC_SIZE];

  assert( qs != NULL );
  // TRD : user_data can be NULL

  queue_internal_new_element_from_freelist( qs, qe, user_data );

  if( qe[QUEUE_POINTER] == NULL )
    return( 0 );

  queue_internal_queue( qs, qe );

  return( 1 );
}





/****************************************************************************/
int queue_guaranteed_enqueue( struct queue_state *qs, void *user_data )
{
  ALIGN(ALIGN_DOUBLE_POINTER) struct queue_element
    *volatile qe[QUEUE_PAC_SIZE];

  assert( qs != NULL );
  // TRD : user_data can be NULL

  queue_internal_guaranteed_new_element_from_freelist( qs, qe, user_data );

  if( qe[QUEUE_POINTER] == NULL )
    return( 0 );

  queue_internal_queue( qs, qe );

  return( 1 );
}





/****************************************************************************/
void queue_internal_queue( struct queue_state *qs, struct queue_element *volatile qe[QUEUE_PAC_SIZE] )
{
  ALIGN(ALIGN_DOUBLE_POINTER) struct queue_element
    *volatile enqueue[QUEUE_PAC_SIZE],
    *volatile original_next[QUEUE_PAC_SIZE];

  unsigned char
    cas_result;

  assert( qs != NULL );
  assert( qe != NULL );

  enqueue[QUEUE_POINTER] = qs->enqueue[QUEUE_POINTER];
  enqueue[QUEUE_COUNTER] = qs->enqueue[QUEUE_COUNTER];

  do
  {
    original_next[QUEUE_POINTER] = NULL;
    original_next[QUEUE_COUNTER] = 0;

    /* TRD : we always have a leading dummy element
             so even when the queue is empty
             enqueue is not NULL and its ->next is safe
    */

    cas_result = abstraction_dcas( (volatile atom_t *) enqueue[QUEUE_POINTER]->next, (atom_t *) qe, (atom_t *) original_next );

    /* TRD : if was NULL, then we just swapped in a nice new element
             (we set the next pointer of the head element to our new element)
             which means head is now pointing at an element which isn't the head
             we check if we can move it along 
             we may not be able to - someone else may have done this already for us
             in which case we don't need to worry about it
    */

    if( cas_result == 1 )
      abstraction_dcas( (volatile atom_t *) qs->enqueue, (atom_t *) qe, (atom_t *) enqueue );

    /* TRD : if was not NULL, then we didn't swap, because someone else has grabbed
             this element before us

             if our enqueue hasn't moved, then someone else has queued a new element
             but hasn't yet moved head; since we know what has to be done, we help them,
             by moving the enqueue point along by one - and then we try again to queue
    */

    if( cas_result == 0 )
      abstraction_dcas( (volatile atom_t *) qs->enqueue, (atom_t *) enqueue[QUEUE_POINTER]->next, (atom_t *) enqueue );
  }
  while( original_next[QUEUE_POINTER] != NULL );

  return;
}





/****************************************************************************/
int queue_dequeue( struct queue_state *qs, void **user_data )
{
  ALIGN(ALIGN_DOUBLE_POINTER) struct queue_element
    *volatile enqueue[QUEUE_PAC_SIZE],
    *volatile dequeue[QUEUE_PAC_SIZE],
    *volatile next[QUEUE_PAC_SIZE];

  unsigned char
    cas_result;

  int
    rv = 1,
    state = FIFO_STATE_UNKNOWN,
    finished_flag = LOWERED;

  assert( qs != NULL );
  assert( user_data != NULL );

  do
  {
    dequeue[QUEUE_POINTER] = qs->dequeue[QUEUE_POINTER];
    dequeue[QUEUE_COUNTER] = qs->dequeue[QUEUE_COUNTER];

    enqueue[QUEUE_POINTER] = qs->enqueue[QUEUE_POINTER];
    enqueue[QUEUE_COUNTER] = qs->enqueue[QUEUE_COUNTER];

    next[QUEUE_POINTER] = dequeue[QUEUE_POINTER]->next[QUEUE_POINTER];
    next[QUEUE_COUNTER] = dequeue[QUEUE_POINTER]->next[QUEUE_COUNTER];

    /* TRD : confirm that dequeue didn't move between reading it
             and reading its next pointer
    */

    if( dequeue[QUEUE_POINTER] == qs->dequeue[QUEUE_POINTER] and dequeue[QUEUE_COUNTER] == qs->dequeue[QUEUE_COUNTER] )
    {
      /* TRD : if enqueue and dequeue are on the same element
               and there is no element in front
               then we're pointing at the dummy element
               it means the queue is empty
      */

      if( enqueue[QUEUE_POINTER] == dequeue[QUEUE_POINTER] and enqueue[QUEUE_COUNTER] == dequeue[QUEUE_COUNTER] and next[QUEUE_POINTER] == NULL )
        state = FIFO_STATE_EMPTY;

      /* TRD : if enqueue and dequeue are on the same element
               and there *is* an element in front
               then the enqueue pointer is behind
               move it along and try again

               although we're dequeuing, we care that the
               enqueue pointer is in the right place because
               if we dequeue this element from out under it,
               all hell will break lose :-)
      */

      if( enqueue[QUEUE_POINTER] == dequeue[QUEUE_POINTER] and enqueue[QUEUE_COUNTER] == dequeue[QUEUE_COUNTER] and next[QUEUE_POINTER] != NULL )
        state = FIFO_STATE_ENQUEUE_OUT_OF_PLACE;

      /* TRD : if enqueue and dequeue differ,
               then maybe enqueue isn't in the right place still
               but it's right *enough* because it's not on us (dequeue)
               so we can safely dequeue
               grab the user_data
               if the dequeue pointer hasn't moved, shift it along
               if it has, go around and try again
      */

      if( enqueue[QUEUE_POINTER] != dequeue[QUEUE_POINTER] or enqueue[QUEUE_COUNTER] != dequeue[QUEUE_COUNTER] )
        state = FIFO_ATTEMPT_DEQUEUE;

      switch( state )
      {
        case FIFO_STATE_EMPTY:
          rv = 0;
          *user_data = NULL;
          finished_flag = RAISED;
        break;

        case FIFO_STATE_ENQUEUE_OUT_OF_PLACE:
          abstraction_dcas( (volatile atom_t *) qs->enqueue, (atom_t *) next, (atom_t *) enqueue );
        break;

        case FIFO_ATTEMPT_DEQUEUE:
          /* TRD : the queue began with a dummy element in the dequeue position
                   we maintain this dummy element by returning
                   the user data from ->next and then dequeuing
                   the dummy element (and moving head forward to ->next)

                   since we have returned the user data from ->next,
                   next is now the new dummy element - still in the queue,
                   but no longer containing user data
          */

          *user_data = next[QUEUE_POINTER]->user_data;

          cas_result = abstraction_dcas( (volatile atom_t *) qs->dequeue, (atom_t *) next, (atom_t *) dequeue );

          if( cas_result == 1 )
            finished_flag = RAISED;
        break;
      }
    }
  }
  while( finished_flag == LOWERED );

  if( state != FIFO_STATE_EMPTY )
    freelist_push( qs->fs, dequeue[QUEUE_POINTER]->fe );

  return( rv );
}

