#include "internal.h"





/****************************************************************************/
void test_queue( void )
{
  printf( "\n"
          "Queue Tests\n"
          "===========\n" );

  queue_test_enqueuing();
  queue_test_dequeuing();
  queue_test_enqueuing_and_dequeuing();
  queue_test_rapid_enqueuing_and_dequeuing();

  return;
}





/****************************************************************************/
void queue_test_enqueuing( void )
{
  unsigned int
    loop,
    cpu_count;

  thread_state_t
    *thread_handles;

  struct queue_state
    *qs;

  struct queue_test_enqueuing_state
    *qtes;

  atom_t
    user_data,
    thread,
    count,
    *per_thread_counters;

  enum error
    e = ERROR_NO_ERROR;

  enum queue_validity
    qv;

  /* TRD : create an empty queue with 1,000,000 elements in its freelist
           then run one thread per CPU
           where each thread busy-works, enqueuing elements (until there are no more elements)
           each element's void pointer of user data is (thread number | element number)
           where element_number is a thread-local counter starting at 0
           where the thread_number occupies the top byte

           when we're done, we check that all the elements are present
           and increment on a per-thread basis
  */

  internal_display_test_name( "Enqueuing" );

  cpu_count = abstraction_cpu_count();

  queue_new( &qs, 1000000 );

  qtes = malloc( sizeof(struct queue_test_enqueuing_state) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    (qtes+loop)->qs = qs;
    (qtes+loop)->thread_number = (atom_t) loop;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], loop, queue_test_internal_thread_simple_enqueuer, qtes+loop );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  free( thread_handles );

  free( qtes );

  /* TRD : first, validate the queue

           then dequeue
           we expect to find element numbers increment on a per thread basis
  */

  queue_query( qs, QUEUE_QUERY_VALIDATE, (void *) (atom_t) 1000000, &qv );

  queue_test_internal_convert_queue_validity_to_error( qv, &e );

  per_thread_counters = malloc( sizeof(atom_t) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    *(per_thread_counters+loop) = 0;

  while( e == ERROR_NO_ERROR and queue_dequeue(qs, (void *) &user_data) )
  {
    thread = user_data >> (sizeof(atom_t)*8-8);
    count = (user_data << 8) >> 8;

    if( thread >= cpu_count )
    {
      e = ERROR_TEST_DATA;
      break;
    }

    if( count < per_thread_counters[thread] )
      e = ERROR_ADDITIONAL_ELEMENTS;

    if( count > per_thread_counters[thread] )
      e = ERROR_MISSING_ELEMENTS;

    if( count == per_thread_counters[thread] )
      per_thread_counters[thread]++;
  }

  free( per_thread_counters );

  queue_delete( qs, NULL, NULL );

  internal_display_test_result( e );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION queue_test_internal_thread_simple_enqueuer( void *queue_test_enqueuing_state )
{
  struct queue_test_enqueuing_state
    *qtes;

  atom_t
    counter;

  assert( queue_test_enqueuing_state != NULL );

  qtes = (struct queue_test_enqueuing_state *) queue_test_enqueuing_state;

  // TRD : we use the top byte for the thread_number
  counter = qtes->thread_number << (sizeof(atom_t)*8-8);

  while( queue_enqueue(qtes->qs, (void *) counter++) );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
void queue_test_dequeuing( void )
{
  unsigned int
    loop,
    cpu_count;

  thread_state_t
    *thread_handles;

  struct queue_state
    *qs;

  struct queue_test_dequeuing_state
    *qtds;

  int
    error_flag = LOWERED;

  /* TRD : create a queue with 1,000,000 elements

           use a single thread to enqueue every element
           each elements user data is an incrementing counter

           then run one thread per CPU
           where each busy-works dequeuing

           when an element is dequeued, we check (on a per-thread basis) the
           value deqeued is greater than the element previously dequeued
  */

  internal_display_test_name( "Dequeuing" );

  cpu_count = abstraction_cpu_count();

  queue_new( &qs, 1000000 );

  for( loop = 0 ; loop < 1000000 ; loop++ )
    queue_enqueue( qs, (void *) (atom_t) loop );

  qtds = malloc( sizeof(struct queue_test_dequeuing_state) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    (qtds+loop)->qs = qs;
    (qtds+loop)->error_flag = LOWERED;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], loop, queue_test_internal_thread_simple_dequeuer, qtds+loop );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  free( thread_handles );

  // TRD : check for raised error flags
  for( loop = 0 ; loop < cpu_count ; loop++ )
    if( (qtds+loop)->error_flag == RAISED )
      error_flag = RAISED;

  free( qtds );

  queue_delete( qs, NULL, NULL );

  if( error_flag == RAISED )
    puts( "failed" );

  if( error_flag == LOWERED )
    puts( "passed" );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION queue_test_internal_thread_simple_dequeuer( void *queue_test_dequeuing_state )
{
  struct queue_test_dequeuing_state
    *qtds;

  atom_t
    *prev_user_data,
    *user_data;

  assert( queue_test_dequeuing_state != NULL );

  qtds = (struct queue_test_dequeuing_state *) queue_test_dequeuing_state;

  queue_dequeue( qtds->qs, (void *) &prev_user_data );

  while( queue_dequeue(qtds->qs, (void *) &user_data) )
  {
    if( user_data < prev_user_data )
      qtds->error_flag = RAISED;

    prev_user_data = user_data;
  }

  return( (thread_return_t) EXIT_SUCCESS );
}




/****************************************************************************/
void queue_test_enqueuing_and_dequeuing( void )
{
  unsigned int
    loop,
    cpu_count;

  thread_state_t
    *thread_handles;

  struct queue_state
    *qs;

  struct queue_test_enqueuing_and_dequeuing_state
    *qteds;

  enum error
    e = ERROR_NO_ERROR;

  atom_t
    user_data,
    thread,
    count,
    *per_thread_counters;

  /* TRD : we have one enqueue and one dequeue thread per CPU
           there is a single queue with 100000 elements
           the enqueue threads busy-work enqueuing,
           where the user-data is the thread number and a counter

           the dequeue threads busy-work dequeuing

           we run for 10 seconds

           after the test, we validate the queue using QUEUE_QUERY_VALIDATE
           then we dequeue fully
           and check counters increment on a per thread basis
  */

  internal_display_test_name( "Enqueuing and dequeuing (10 seconds)" );

  cpu_count = abstraction_cpu_count();

  queue_new( &qs, 100000 );

  qteds = malloc( sizeof(struct queue_test_enqueuing_and_dequeuing_state) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    (qteds+loop)->qs = qs;
    // TRD : init top byte of counter with thread number
    (qteds+loop)->counter = (atom_t) loop << (sizeof(atom_t)*8-8);
    (qteds+loop)->thread_number = (atom_t) loop;
  }

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 2 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], loop, queue_test_internal_thread_enqueuer, qteds+loop );
    abstraction_thread_start( &thread_handles[loop+cpu_count], loop, queue_test_internal_thread_dequeuer, qs );
  }

  for( loop = 0 ; loop < cpu_count * 2 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  free( thread_handles );

  /* TRD : dequeue and check user data
           each struct queue_test_enqueuing_and_dequeuing_state contains
           the number of queues performed by that thread;

           that number is the value we expect to find as the final user
           data counter value for that thread

           so we track that on a per-thread basis
           they always increment and we check that the final values we
           see matches the final values we expect
  */

  per_thread_counters = malloc( sizeof(atom_t) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    *(per_thread_counters+loop) = 0;

  while( e == ERROR_NO_ERROR and queue_dequeue(qs, (void *) &user_data) )
  {
    thread = user_data >> (sizeof(atom_t)*8-8);
    count = (user_data << 8) >> 8;

    if( thread >= cpu_count )
    {
      e = ERROR_TEST_DATA;
      break;
    }

    if( per_thread_counters[thread] == 0 )
      per_thread_counters[thread] = count;

    if( count < per_thread_counters[thread] )
      e = ERROR_ADDITIONAL_ELEMENTS;

    if( count > per_thread_counters[thread] )
      e = ERROR_MISSING_ELEMENTS;

    if( count == per_thread_counters[thread] )
      per_thread_counters[thread]++;
  }

  // TRD : now check the final values match up
  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    /* TRD : if per_thread_counters[thread] is 0, it means that thread
             did no work that remains present in the queue
    */

    if( per_thread_counters[loop] != 0 and per_thread_counters[loop] != (((qteds+loop)->counter << 8) >> 8) )
      e = ERROR_TEST_DATA;
  }

  free( per_thread_counters );

  free( qteds );

  queue_delete( qs, NULL, NULL );

  internal_display_test_result( e );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION queue_test_internal_thread_enqueuer( void *queue_test_enqueuing_and_dequeuing_state )
{
  struct queue_test_enqueuing_and_dequeuing_state
    *qteds;

  time_t
    start_time;

  assert( queue_test_enqueuing_and_dequeuing_state != NULL );

  qteds = (struct queue_test_enqueuing_and_dequeuing_state *) queue_test_enqueuing_and_dequeuing_state;

  time( &start_time );

  while( time(NULL) < start_time + 10 )
    if( queue_enqueue(qteds->qs, (void *) qteds->counter) )
      qteds->counter++;

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION queue_test_internal_thread_dequeuer( void *queue_state )
{
  struct queue_state
    *qs;

  time_t
    start_time;

  void
    *user_data;

  assert( queue_state != NULL );

  qs = (struct queue_state *) queue_state;

  time( &start_time );

  while( time(NULL) < start_time + 10 )
    queue_dequeue( qs, &user_data );

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
void queue_test_rapid_enqueuing_and_dequeuing( void )
{
  unsigned int
    loop,
    cpu_count;

  thread_state_t
    *thread_handles;

  struct queue_state
    *qs;

  enum error
    e = ERROR_NO_ERROR;

  enum queue_validity
    qv;

  internal_display_test_name( "Rapid enqueuing and dequeuing (10 seconds)" );

  cpu_count = abstraction_cpu_count();

  queue_new( &qs, 1000 );

  for( loop = 0 ; loop < 500 ; loop++ )
    queue_enqueue( qs, NULL );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], loop, queue_test_internal_thread_rapid_enqueuer_and_dequeuer, qs );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  free( thread_handles );

  queue_query( qs, QUEUE_QUERY_VALIDATE, (void *) (atom_t) 500, &qv );

  queue_test_internal_convert_queue_validity_to_error( qv, &e );

  queue_delete( qs, NULL, NULL );

  internal_display_test_result( e );

  return;
}





/****************************************************************************/
thread_return_t CALLING_CONVENTION queue_test_internal_thread_rapid_enqueuer_and_dequeuer( void *queue_state )
{
  struct queue_state
    *qs;

  time_t
    start_time;

  void
    *user_data = NULL;

  assert( queue_state != NULL );

  qs = (struct queue_state *) queue_state;

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    queue_enqueue( qs, user_data );
    queue_dequeue( qs, &user_data );
  }

  return( (thread_return_t) EXIT_SUCCESS );
}





/****************************************************************************/
void queue_test_internal_convert_queue_validity_to_error( enum queue_validity qv, enum error *e )
{
  // TRD : qv can be any value in its range
  assert( e != NULL );

  /* TRD : enum error is a superset of all the individual query_validity enums
           so we never need to worry about unrepresentable data
  */

  switch( qv )
  {
    case QUEUE_VALID:
      *e = ERROR_NO_ERROR;
    break;

    case QUEUE_INVALID_LOOP:
      *e = ERROR_LOOP;
    break;

    case QUEUE_INVALID_MISSING_ELEMENTS:
      *e = ERROR_MISSING_ELEMENTS;
    break;

    case QUEUE_INVALID_ADDITIONAL_ELEMENTS:
      *e = ERROR_ADDITIONAL_ELEMENTS;
    break;
  }

  return;
}

